/*****************************************************************************
 * atmo.c : "Atmo Light" video filter
 *****************************************************************************
 * Copyright (C) 2000-2006 the VideoLAN team
 * $Id$
 *
 * Authors: Andr Weber (WeberAndre@gmx.de)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#include <stdlib.h>                                      /* malloc(), free() */
#include <string.h>
#include <math.h>                                            /* sin(), cos() */

#include <vlc/vlc.h>
#include <vlc/vout.h>
#include <vlc_input.h>
#include <vlc_playlist.h>
#include "filter_common.h"
#include "vlc_image.h"

/*****************************************************************************
 * Local prototypes
 *****************************************************************************/
static int  CreateFilter    ( vlc_object_t * );
static void DestroyFilter   ( vlc_object_t * );

static int InitFilter( vout_thread_t *p_vout );
static void EndFilter( vout_thread_t *p_vout );

static void Render( vout_thread_t *p_vout, picture_t *p_pic );
static int Control( vout_thread_t *p_vout, int i_query, va_list args );
static int SendEvents( vlc_object_t *, char const *, vlc_value_t, vlc_value_t, void * ); 
static int SendEventsToChild( vlc_object_t *, char const *, vlc_value_t, vlc_value_t, void *); 

static int StateCallback(vlc_object_t *p_this, char const *psz_cmd,vlc_value_t oldval, vlc_value_t newval, void *p_data);

static void CheckAndStopFadeThread(vout_thread_t *p_vout);

/* what was first vout_sys_t or my function chicken egg problem?*/
typedef void (ExtractMiniImage(/*vout_sys_t*/void *p_sys, picture_t *p_inpic, uint8_t *p_transfer_dest));


/*****************************************************************************
* External Prototypes for the AtmoCtrlLib.DLL 
*****************************************************************************/
/* 
* if effectmode = emLivePicture then the source could be GDI (Screencapture) 
* or External - this means another application delivers Pixeldata to AtmoWin
* Clientsoftware through  AtmoCtrlLib.DLL and the COM Api
*/
#define lvsGDI	         0
#define lvsExternal      1

/*
* current Effect mode the only one used here is emLivePicture the other #define
* are currently not used in this code...
*/
#define emDisabled	     0
#define	emStaticColor    1
#define	emLivePicture    2
#define emColorChange    3
#define	emLrColorChange  4


/*******************************************************************************
 * Function prototypes for COM wrapper DLL because I wasn't able to get the COM 
 * stuff compiled under cygwin and MingW -- so I moved all the COM calls to another 
 * DLL which could be compiled with Visual Studio... 
 *
 * Sources for this DLL and the ClientSoftware (AtmoWin) which controls the
 * hardware could be found here - it's free 
 * http://eldo.gotdns.com/atmo/ 
 * http://www.vdr-portal.de/board/thread.php?threadid=59294
********************************************************************************/
typedef int (AtmoInitialize(void));
typedef void (AtmoFinalize(int what));
typedef int (AtmoSwitchEffect(int newMode));
typedef int (AtmoSetLiveSource(int newSource));
typedef void (AtmoCreateTransferBuffers(int FourCC,int bytePerPixel,int width,int height));
typedef uint8_t* (AtmoLockTransferBuffer(void));
typedef void (AtmoSendPixelData(void));


/*
 strings for settings menus and hints
*/
#define SAVEFRAMES_TEXT N_("Save Debug Frames")
#define SAVEFRAMES_LONGTEXT N_("Writes every 128th miniframe to a folder.")

#define FRAMEPATH_TEXT N_("Debug Frame Folder")
#define FRAMEPATH_LONGTEXT N_("defines the path where the debugframes should be saved")

#define WIDTH_TEXT N_("Extracted Image Width")
#define WIDTH_LONGTEXT N_("defines the width of the mini image for further processing (64 is default)")

#define HEIGHT_TEXT N_("Extracted Image Height")
#define HEIGHT_LONGTEXT N_("defines the height of the mini image for further processing (48 is default)")

#define PCOLOR_TEXT N_("use Pause Color")
#define PCOLOR_LONGTEXT N_("use the color defined below if the user paused the video.(have light to get another beer?)")

#define PCOLOR_RED_TEXT N_("Red")
#define PCOLOR_RED_LONGTEXT N_("the red component of pause color")
#define PCOLOR_GREEN_TEXT N_("Green")
#define PCOLOR_GREEN_LONGTEXT N_("the green component of pause color")
#define PCOLOR_BLUE_TEXT N_("Blue")
#define PCOLOR_BLUE_LONGTEXT N_("the blue component of pause color")

#define ATMOWINEXE_TEXT N_("Filename of AtmoWin.exe")
#define ATMOWINEXE_LONGTEXT N_("if you wan't that the AtmoLight control software is launched by\nVLC enter the complete Filename of AtmoWin.exe here")
#define FADESTEPS_TEXT N_("Fadesteps")
#define FADESTEPS_LONGTEXT N_("Number of steps to change current color to pause color (each step takes 40ms)")


/*****************************************************************************
 * Module descriptor
 *****************************************************************************/
vlc_module_begin();
    set_description( _("AtmoLight Filter") );
    set_shortname( _( "AtmoLight" ));
    set_capability( "video filter", 0 );
    set_category( CAT_VIDEO );
    set_subcategory( SUBCAT_VIDEO_VFILTER );

    add_bool("atmo-usepausecolor", VLC_FALSE, NULL, PCOLOR_TEXT, PCOLOR_LONGTEXT, VLC_FALSE);
    add_integer_with_range("atmo-pcolor-red",   128, 0, 255, NULL, PCOLOR_RED_TEXT,   PCOLOR_RED_LONGTEXT,   VLC_FALSE);
    add_integer_with_range("atmo-pcolor-green", 128, 0, 255, NULL, PCOLOR_GREEN_TEXT, PCOLOR_GREEN_LONGTEXT, VLC_FALSE);
    add_integer_with_range("atmo-pcolor-blue",  128, 0, 255, NULL, PCOLOR_BLUE_TEXT,  PCOLOR_BLUE_LONGTEXT,  VLC_FALSE);
    add_integer_with_range("atmo-fadesteps",     75, 1, 250, NULL, FADESTEPS_TEXT,    FADESTEPS_LONGTEXT,    VLC_FALSE);

    add_string("atmo-atmowinexe", "", NULL, ATMOWINEXE_TEXT, ATMOWINEXE_LONGTEXT, VLC_FALSE );

    add_bool("atmo-saveframes", VLC_FALSE, NULL, SAVEFRAMES_TEXT, SAVEFRAMES_LONGTEXT, VLC_FALSE);
    add_string("atmo-framepath", "", NULL, FRAMEPATH_TEXT, FRAMEPATH_LONGTEXT, VLC_FALSE );

    add_integer_with_range("atmo-width",  64, 64, 512, NULL, WIDTH_TEXT,  WIDTH_LONGTEXT, VLC_TRUE);
    add_integer_with_range("atmo-height", 48, 48, 384, NULL, HEIGHT_TEXT,  HEIGHT_LONGTEXT, VLC_TRUE);

    add_shortcut( "atmo" );
    set_callbacks( CreateFilter, DestroyFilter  );
vlc_module_end();

/*****************************************************************************
 * fadethread_t: Color Fading Thread
 *****************************************************************************
 * changes slowly the color of the output if videostream gets paused...
 *****************************************************************************
*/
typedef struct  
{
    VLC_COMMON_MEMBERS
    void *p_sys;
} fadethread_t;

/*****************************************************************************
 * vout_sys_t: AtmoLight output method descriptor 
 *****************************************************************************
 * This structure is part of the video output thread descriptor.
 * It describes the AtmoLight specific properties of an output thread.
 *****************************************************************************/
struct vout_sys_t
{
    vout_thread_t *p_vout;
    /* special for the access of the p_fadethread member all other members need no special protection so far! */
    vlc_mutex_t filter_lock;

    int i_enabled;
    int i_AtmoOldEffect;
    vlc_bool_t b_pause_live;

    int i_atmo_width;
    int i_atmo_height;

    vlc_bool_t  b_saveframes;
    int i_framecounter;
    char sz_framepath[MAX_PATH];

    vlc_bool_t  b_usepausecolor;
    uint8_t ui_pausecolor_red;
    uint8_t ui_pausecolor_green;
    uint8_t ui_pausecolor_blue;
    int i_fadesteps;
    fadethread_t *p_fadethread;

    ExtractMiniImage *pf_ExtractMiniImage;

#if defined( WIN32 )
    /* External Library as wrapper arround my COM Stuff */
    HINSTANCE h_AtmoCtrl;
#endif

    AtmoInitialize *pf_AtmoInitialize;
    AtmoFinalize *pf_AtmoFinalize;
    AtmoSwitchEffect *pf_AtmoSwitchEffect;
    AtmoSetLiveSource *pf_AtmoSetLiveSource;
    AtmoCreateTransferBuffers *pf_AtmoCreateTransferBuffers;
    AtmoLockTransferBuffer *pf_AtmoLockTransferBuffer;
    AtmoSendPixelData *pf_AtmoSendPixelData;
};

#if !defined( WIN32 )
/* dummy functions for non win32 environments */

static int nonwin32_AtmoInitialize(void) {
       /* return 0 => failure! */
       return 0; 
}
static void nonwin32_AtmoFinalize(int what) {
} 

static int nonwin32_AtmoSwitchEffect(int newMode) {
       return emDisabled;
}

static int nonwin32_AtmoSetLiveSource(int newSource) {
       return lvsGDI;
}

static void nonwin32_AtmoCreateTransferBuffers(int FourCC,int bytePerPixel,int width,int height) {
       // get mutex for global buffer
       // if(p_non_windowstransferbuffer != NULL)
       //   free(p_non_windowstransferbuffer);
       // p_non_windowstransferbuffer = (uint8_t *)malloc( width * height * bytePerPixel );
       // release mutex
}
static uint8_t* nonwin32_AtmoLockTransferBuffer(void) {
       // get mutex for global buffer variable?
       return NULL;
}

static void nonwin32_AtmoSendPixelData(void) {
       // transfer data - to another thread which does conversation RGB -- HSV 
       // and the statistics...
       // release mutex locked by AtmoLockTransferBuffer 
}

#endif


/*****************************************************************************
 * CreateFilter: allocates AtmoLight video thread output method
 *****************************************************************************
 * This function allocates and initializes a AtmoLight vout method.
 *****************************************************************************/
static int CreateFilter( vlc_object_t *p_this )
{
    vout_thread_t *p_vout = (vout_thread_t *)p_this;
    vout_sys_t *p_sys;

    /* Allocate structure */
    p_sys = malloc( sizeof( vout_sys_t ) );;
    p_vout->p_sys = p_sys;
    if( p_vout->p_sys == NULL )
    {
        msg_Err( p_vout, "out of memory for p_sys structure" );
        return VLC_ENOMEM;
    }
    /* set all entries to zero */
    memset(p_sys, 0, sizeof( vout_sys_t ));

    /* default filter disabled until DLL loaded and Init Success!*/
    p_sys->i_enabled       = 0;

    /* setup default mini image size (may be later a user option) */
    p_sys->i_atmo_width    = 64;
    p_sys->i_atmo_height   = 48;
    p_sys->i_framecounter  = 0;
    p_sys->i_fadesteps     = 1;
    p_sys->pf_ExtractMiniImage = NULL;
    vlc_mutex_init( p_vout, &p_sys->filter_lock );


#if defined( WIN32 )
    /* Load the Com Wrapper Library (source available) */
    p_sys->h_AtmoCtrl = LoadLibraryA("AtmoCtrlLib.dll");
    if(p_sys->h_AtmoCtrl != NULL) 
    {
       msg_Dbg( p_vout, "LoadLibrary('AtmoCtrlLib.dll'); Success");  

       /* importing all required functions I hope*/
       p_sys->pf_AtmoInitialize            = (AtmoInitialize *)GetProcAddress(p_sys->h_AtmoCtrl,"AtmoInitialize");
       if(!p_sys->pf_AtmoInitialize)
          msg_Err( p_vout, "export AtmoInitialize missing."); 

       p_sys->pf_AtmoFinalize              = (AtmoFinalize *)GetProcAddress(p_sys->h_AtmoCtrl,"AtmoFinalize");
       if(!p_sys->pf_AtmoFinalize)
          msg_Err( p_vout, "export AtmoFinalize missing."); 

       p_sys->pf_AtmoSwitchEffect          = (AtmoSwitchEffect *)GetProcAddress(p_sys->h_AtmoCtrl,"AtmoSwitchEffect");
       if(!p_sys->pf_AtmoSwitchEffect)
          msg_Err( p_vout, "export AtmoSwitchEffect missing."); 

       p_sys->pf_AtmoSetLiveSource         = (AtmoSetLiveSource *)GetProcAddress(p_sys->h_AtmoCtrl,"AtmoSetLiveSource");
       if(!p_sys->pf_AtmoSetLiveSource)
          msg_Err( p_vout, "export AtmoSetLiveSource missing."); 

       p_sys->pf_AtmoCreateTransferBuffers = (AtmoCreateTransferBuffers *)GetProcAddress(p_sys->h_AtmoCtrl,"AtmoCreateTransferBuffers");
       if(!p_sys->pf_AtmoCreateTransferBuffers)
          msg_Err( p_vout, "export AtmoCreateTransferBuffers missing."); 

       p_sys->pf_AtmoLockTransferBuffer    = (AtmoLockTransferBuffer *)GetProcAddress(p_sys->h_AtmoCtrl,"AtmoLockTransferBuffer");
       if(!p_sys->pf_AtmoLockTransferBuffer)
          msg_Err( p_vout, "export AtmoLockTransferBuffer missing."); 

       p_sys->pf_AtmoSendPixelData         = (AtmoSendPixelData *)GetProcAddress(p_sys->h_AtmoCtrl,"AtmoSendPixelData");
       if(!p_sys->pf_AtmoSendPixelData)
          msg_Err( p_vout, "export AtmoSendPixelData missing."); 
    }
#else
    /*
      for non win32 environments use these dummy function pointers
      which do really nothing to avoid crashing the vlc on loading this
      module - may be somebody is willing to implement it for linux
      he can make use of this?
    */
    p_sys->pf_AtmoInitialize     = nonwin32_AtmoInitialize;
    p_sys->pf_AtmoFinalize       = nonwin32_AtmoFinalize;
    p_sys->pf_AtmoSwitchEffect   = nonwin32_AtmoSwitchEffect;
    p_sys->pf_AtmoSetLiveSource  = nonwin32_AtmoSetLiveSource;
    p_sys->pf_AtmoCreateTransferBuffers = nonwin32_AtmoCreateTransferBuffers;
    p_sys->pf_AtmoLockTransferBuffer = nonwin32_AtmoLockTransferBuffer;
    p_sys->pf_AtmoSendPixelData = nonwin32_AtmoSendPixelData;
#endif

   
   
    /* further Setup Function pointers for videolan for calling my filter */
    p_vout->pf_init    = InitFilter;
    p_vout->pf_end     = EndFilter;
    p_vout->pf_manage  = NULL;
    p_vout->pf_render  = Render;
    p_vout->pf_display = NULL;
    p_vout->pf_control = Control;
    
    return VLC_SUCCESS;
}


/*****************************************************************************
 * DestroyFilter: destroy AtmoLight video thread output method
 *****************************************************************************
 * Terminate an output method created by CreateFilter
 *****************************************************************************/
static void DestroyFilter( vlc_object_t *p_this )
{
    vout_thread_t *p_vout = (vout_thread_t *)p_this;
    vout_sys_t *p_sys =  p_vout->p_sys;


#if defined( WIN32 )
    if(p_sys->h_AtmoCtrl!=NULL) 
    {
       FreeLibrary(p_sys->h_AtmoCtrl);
    }
#endif

    vlc_mutex_destroy( &p_sys->filter_lock );
    free( p_sys );
}




/*
 function stolen from some other videolan source filter ;-)
 for the moment RGB is OK... but better would be a direct transformation
 from YUV --> HSV 
*/
static inline void yuv_to_rgb( uint8_t *r, uint8_t *g, uint8_t *b,
                               uint8_t y1, uint8_t u1, uint8_t v1 )
{
    /* macros used for YUV pixel conversions */
#   define SCALEBITS 10
#   define ONE_HALF  (1 << (SCALEBITS - 1))
#   define FIX(x)    ((int) ((x) * (1<<SCALEBITS) + 0.5))
#   define CLAMP( x ) (((x) > 255) ? 255 : ((x) < 0) ? 0 : (x));

    int y, cb, cr, r_add, g_add, b_add;

    cb = u1 - 128;
    cr = v1 - 128;
    r_add = FIX(1.40200*255.0/224.0) * cr + ONE_HALF;
    g_add = - FIX(0.34414*255.0/224.0) * cb
            - FIX(0.71414*255.0/224.0) * cr + ONE_HALF;
    b_add = FIX(1.77200*255.0/224.0) * cb + ONE_HALF;
    y = (y1 - 16) * FIX(255.0/219.0);
    *r = CLAMP((y + r_add) >> SCALEBITS);
    *g = CLAMP((y + g_add) >> SCALEBITS);
    *b = CLAMP((y + b_add) >> SCALEBITS);
}
/***********************************************************************************
 * ExtractMiniImage_YUV: extract a small image from the picture as 24-bit RGB
 ***********************************************************************************
 * p_system is a pointer to vout_sys_t
 * p_inpic is the source frame
 * p_transfer_dest is the target buffer for the picture must be big enough!
 * (in win32 enviroment this buffer comes from the external DLL where it is
 * create as "variant array" and returned through the AtmoLockTransferbuffer
*/
static void ExtractMiniImage_YUV(void *p_system, picture_t *p_inpic, uint8_t *p_transfer_dest) 
{
    vout_sys_t *p_sys = (vout_sys_t *)p_system;
    int i_col;
    int i_row;
    uint8_t *p_src_y;
    uint8_t *p_src_u;
    uint8_t *p_src_v;
    uint8_t *p_rgb_dst_line_red;
    uint8_t *p_rgb_dst_line_green;
    uint8_t *p_rgb_dst_line_blue;
    int i_xpos_y;
    int i_xpos_u;
    int i_xpos_v;

    /* calcute Pointers for Storage of B G R (A) */
    p_rgb_dst_line_blue      = p_transfer_dest;
    p_rgb_dst_line_green     = p_transfer_dest + 1;
    p_rgb_dst_line_red       = p_transfer_dest + 2 ;

    int i_row_count = p_sys->i_atmo_height + 1;
    int i_col_count = p_sys->i_atmo_width + 1;

    /*  these two ugly loops extract the small image - goes it faster? how?
        the loops are so designed that there is a small border around the extracted image
        so we wont get column and row - zero from the frame, and not the most right 
        and bottom pixels --- which may be clipped on computers useing TV out - through
        overscan!

        TODO: try to find out if the output is clipped through VLC - and try here
        to ingore the clipped away area for a better result!

        TODO: performance improvement in InitFilter percalculated the offsets of
        the lines inside the planes so I can save (i_row_count * 3) 2xMUL and one time DIV 
        the same could be done for the inner loop I think... 
    */
    for(i_row = 1; i_row < i_row_count; i_row++) 
    {
        // calcute the current Lines in the source planes for this outputrow
        p_src_y = p_inpic->p[Y_PLANE].p_pixels + p_inpic->p[Y_PLANE].i_pitch * ((i_row * p_inpic->p[Y_PLANE].i_visible_lines)/i_row_count);
        p_src_u = p_inpic->p[U_PLANE].p_pixels + p_inpic->p[U_PLANE].i_pitch * ((i_row * p_inpic->p[U_PLANE].i_visible_lines)/i_row_count);
        p_src_v = p_inpic->p[V_PLANE].p_pixels + p_inpic->p[V_PLANE].i_pitch * ((i_row * p_inpic->p[V_PLANE].i_visible_lines)/i_row_count);

        for(i_col = 1; i_col < i_col_count; i_col++) 
        {
            i_xpos_y = (p_inpic->p[Y_PLANE].i_visible_pitch * i_col) / i_col_count; 
            i_xpos_u = (p_inpic->p[U_PLANE].i_visible_pitch * i_col) / i_col_count; 
            i_xpos_v = (p_inpic->p[V_PLANE].i_visible_pitch * i_col) / i_col_count; 

            yuv_to_rgb(p_rgb_dst_line_red, p_rgb_dst_line_green, p_rgb_dst_line_blue, p_src_y[i_xpos_y], p_src_u[i_xpos_u], p_src_v[i_xpos_v]);
            
            /* +4 because output image should be RGB32 with dword alignment! */
            p_rgb_dst_line_red   += 4;
            p_rgb_dst_line_green += 4;
            p_rgb_dst_line_blue  += 4;
        }
    }
}


/************************************************************************************************
 * AddPlaylistInputThreadStateCallback: Setup call back on "State" Variable
 ************************************************************************************************
 * Add Callback function to the "state" variable of the input thread.. 
 * first find the PlayList and get the input thread from there to attach
 * my callback? is vlc_object_find the right way for this??
 ***********************************************************************************************/
static void AddPlaylistInputThreadStateCallback(vout_thread_t *p_vout) 
{
 playlist_t *p_playlist = (playlist_t *)vlc_object_find( p_vout, VLC_OBJECT_PLAYLIST, FIND_ANYWHERE );
 if( p_playlist )
  {
    input_thread_t *p_input = p_playlist->p_input;
    if(p_input) 
    {
       var_AddCallback( p_input, "state", StateCallback, p_vout );
    }
    vlc_object_release( p_playlist );
  }
}

/************************************************************************************************
 * DelPlaylistInputThreadStateCallback: Remove call back on "State" Variable
 ************************************************************************************************
 * Delete the callback function to the "state" variable of the input thread.. 
 * first find the PlayList and get the input thread from there to attach
 * my callback? is vlc_object_find the right way for this??
 ***********************************************************************************************/
static void DelPlaylistInputThreadStateCallback( vout_thread_t *p_vout ) 
{
 playlist_t *p_playlist = (playlist_t *)vlc_object_find( p_vout, VLC_OBJECT_PLAYLIST, FIND_ANYWHERE );
 if( p_playlist )
 {
    input_thread_t *p_input = p_playlist->p_input;
    if(p_input) 
    {
       var_DelCallback( p_input, "state", StateCallback, p_vout );
    }
    vlc_object_release( p_playlist );
  }
}

/*****************************************************************************
 * InitFilter: initialize AtmoLight video thread 
 *****************************************************************************/
static int InitFilter( vout_thread_t *p_vout )
{
    int i_index;
    char *psz_path;
    picture_t *p_pic;
    video_format_t fmt = {0};

    vout_sys_t *p_sys = p_vout->p_sys;

    I_OUTPUTPICTURES = 0;

    /* Initialize the output structure */
    p_vout->output.i_chroma = p_vout->render.i_chroma;
    p_vout->output.i_width  = p_vout->render.i_width;
    p_vout->output.i_height = p_vout->render.i_height;
    p_vout->output.i_aspect = p_vout->render.i_aspect;
    p_vout->fmt_out = p_vout->fmt_in;
    fmt = p_vout->fmt_out;
    msg_Dbg( p_vout, "InitFilter-Chroma: %4.4s",(char *)&p_vout->render.i_chroma);

    switch( p_vout->render.i_chroma )
    {
    case VLC_FOURCC('I','4','2','0'):
         p_sys->pf_ExtractMiniImage = ExtractMiniImage_YUV;
    break;
    case VLC_FOURCC('I','Y','U','V'):
         p_sys->pf_ExtractMiniImage = ExtractMiniImage_YUV;
    break;
    case VLC_FOURCC('Y','V','1','2'):
         p_sys->pf_ExtractMiniImage = ExtractMiniImage_YUV;
    break;
    case VLC_FOURCC('Y','V','1','6'):
         p_sys->pf_ExtractMiniImage = ExtractMiniImage_YUV;
    break;
    case VLC_FOURCC('Y','V','U','9'):
         p_sys->pf_ExtractMiniImage = ExtractMiniImage_YUV; 
    break;

    default:
        msg_Dbg( p_vout, "InitFilter-unsupported chroma: %4.4s",(char *)&p_vout->render.i_chroma);
        return VLC_EGENERIC; /* unknown chroma */
    }

    /* Try to open the real video output */
    msg_Dbg( p_vout, "InitFilter-spawn AtmoLight Filter Thread" );

    p_sys->p_vout = vout_Create( p_vout, &fmt );
    /* Everything failed */
    if( p_sys->p_vout == NULL )
    {
        msg_Err( p_vout, "cannot open vout for AtmoLight, aborting" );
        return VLC_EGENERIC;
    }

    AddPlaylistInputThreadStateCallback(p_vout);

    ALLOCATE_DIRECTBUFFERS( VOUT_MAX_PICTURES );
    ADD_CALLBACKS( p_sys->p_vout, SendEvents );
    ADD_PARENT_CALLBACKS( SendEventsToChild );

    /* save debug images to a folder as Bitmap files ? */
    p_sys->b_saveframes  = var_CreateGetBool( p_vout, "atmo-saveframes" );
    msg_Dbg(p_vout,"saveframes = %d", (int)p_sys->b_saveframes);

    /* read debug image folder from config */
    psz_path = config_GetPsz( p_vout, "atmo-framepath" );
    if(psz_path != NULL) 
    {
        strcpy(p_sys->sz_framepath, psz_path);
#if defined( WIN32 )
        int i_strlen = strlen(p_sys->sz_framepath);
        if((i_strlen>0) && (p_sys->sz_framepath[i_strlen-1] != '\\')) 
        {
           p_sys->sz_framepath[i_strlen] = '\\';
           p_sys->sz_framepath[i_strlen+1] = 0;
        }
#endif
       free(psz_path);
    }
    msg_Dbg(p_vout,"saveframesfolder %s",p_sys->sz_framepath);

    /* size of extracted image by default 64x48 (other imagesizes are currently ignored by AtmoWin)  */
    p_sys->i_atmo_width  = var_CreateGetInteger( p_vout, "atmo-width");
    p_sys->i_atmo_height = var_CreateGetInteger( p_vout, "atmo-height");
    msg_Dbg(p_vout,"mini image size %d * %d pixels", p_sys->i_atmo_width, p_sys->i_atmo_height);

    /* 
    because atmowin could also be used for lighten up the room - I think if you
    pause the video it would be useful to get a little bit more light into to
    your living room? - instead switching on a lamp?
    */
    p_sys->b_usepausecolor      = var_CreateGetBool( p_vout, "atmo-usepausecolor" );
    p_sys->ui_pausecolor_red    = (uint8_t)var_CreateGetInteger( p_vout, "atmo-pcolor-red");
    p_sys->ui_pausecolor_green  = (uint8_t)var_CreateGetInteger( p_vout, "atmo-pcolor-green");
    p_sys->ui_pausecolor_blue   = (uint8_t)var_CreateGetInteger( p_vout, "atmo-pcolor-blue"); 
    p_sys->i_fadesteps          = var_CreateGetInteger( p_vout, "atmo-fadesteps"); 
    if(p_sys->i_fadesteps < 1)
       p_sys->i_fadesteps = 1;
    msg_Dbg(p_vout,"use pause color %d, RGB: %d, %d, %d, Fadesteps: %d",(int)p_sys->b_usepausecolor, p_sys->ui_pausecolor_red, p_sys->ui_pausecolor_green, p_sys->ui_pausecolor_blue, p_sys->i_fadesteps);
    p_sys->b_pause_live         = VLC_FALSE; 

    /* if the external DLL was loaded successfully call AtmoInitialize - (must be done for each thread where you wan't to use AtmoLight! */
    if(p_sys->pf_AtmoInitialize != NULL) 
    {
        int i = p_sys->pf_AtmoInitialize();
        if(i != 1) 
        {
          /* COM Server for AtmoLight not running ? 
             if the exe path is configured try to start the "userspace" driver
          */
          psz_path = config_GetPsz( p_vout, "atmo-atmowinexe" );
          if(psz_path != NULL) 
          { 
#if defined( WIN32 )

             STARTUPINFO startupinfo;
             PROCESS_INFORMATION pinfo;
             memset(&startupinfo, 0, sizeof(STARTUPINFO));
             startupinfo.cb = sizeof(STARTUPINFO);
             if(CreateProcess(psz_path, NULL, NULL, NULL, FALSE, 0, NULL, NULL, &startupinfo, &pinfo) == TRUE) 
             {
                 msg_Dbg(p_vout,"launched AtmoWin from %s",psz_path);
                 WaitForInputIdle(pinfo.hProcess, 5000);
                 // retry to initialize the library COM ... functionality after the server was launched
                 i = p_sys->pf_AtmoInitialize(); 
             } else {
                 msg_Err(p_vout,"failed to launch AtmoWin from %s",psz_path);
             }
#endif
             free(psz_path);
          }
        }

        if(i == 1) /* Init Atmolight success... */
        {
            msg_Dbg( p_vout, "AtmoInitialize Ok!");

            /* Setup Transferbuffers for 64 x 48 , RGB with 32bit Per Pixel */
            p_sys->pf_AtmoCreateTransferBuffers(BI_RGB, 4, p_sys->i_atmo_width, p_sys->i_atmo_height);

            /* say the userspace driver that a live mode should be activated 
               the functions returns the old mode for later restore!
            */
            p_sys->i_AtmoOldEffect = p_sys->pf_AtmoSwitchEffect(emLivePicture);

            /* 
                live view can have two differnt source the AtmoWinA internal GDI Screencapture
                and the external one - which we need here...
            */
            p_sys->pf_AtmoSetLiveSource(lvsExternal);

            /* enable other parts only if everything is fine */
            p_sys->i_enabled = 1; 
        }  
    }

    return VLC_SUCCESS;
}


/*****************************************************************************
 * EndFilter: terminate AtmoLight video thread output method
 *****************************************************************************/
static void EndFilter( vout_thread_t *p_vout )
{
    int i_index;
    vout_sys_t *p_sys = p_vout->p_sys;

    /* Free the fake output buffers we allocated */
    for( i_index = I_OUTPUTPICTURES ; i_index ; )
    {
        i_index--;
        free( PP_OUTPUTPICTURE[ i_index ]->p_data_orig );
    }

    DelPlaylistInputThreadStateCallback(p_vout);

    if( p_sys->p_vout )
    {
        DEL_CALLBACKS( p_sys->p_vout, SendEvents );
        vlc_object_detach( p_sys->p_vout );
        vout_Destroy( p_sys->p_vout );
    }
    DEL_PARENT_CALLBACKS( SendEventsToChild );


    if(p_sys->i_enabled == 1) 
    {
       CheckAndStopFadeThread(p_vout);

       if(p_sys->pf_AtmoFinalize != NULL)
       {
           msg_Dbg( p_vout, "EndFilter Switchback to AtmoMode %d", p_sys->i_AtmoOldEffect);

           /* restore AtmoWinA Mode .. to old effect before VideoLan! */
          if(p_sys->i_AtmoOldEffect != emLivePicture)
             p_sys->pf_AtmoSwitchEffect(p_sys->i_AtmoOldEffect);

          if(p_sys->i_AtmoOldEffect == emLivePicture)
             p_sys->pf_AtmoSetLiveSource(lvsGDI);

          /* cleanup Library Globals VarArrays, Global COM Object pointers (Release them) 
             and call OleUnitialize (on win32)
             1 => vararray and ole
          */
          p_sys->pf_AtmoFinalize(1);
       }

       /* disable filter method .. */
       p_sys->i_enabled = 0;
    }


}

/******************************************************************************
* SaveBitmap: Saves the content of a transferbuffer as Bitmap to disk
*******************************************************************************
* just for debugging 
* p_sys -> configuration if Atmo from there the function will get height and width
* p_pixels -> should be the dword aligned BGR(A) image data
* psz_filename -> filename where to store 
*/
void SaveBitmap(vout_sys_t *p_sys, uint8_t *p_pixels, char *psz_filename) 
{
#if defined( WIN32 )
    /* for debug out only used*/
    BITMAPINFO bmp_info;
    BITMAPFILEHEADER  bmp_fileheader; 
    FILE *fp_bitmap;

    memset(&bmp_info, 0, sizeof(BITMAPINFO));
    bmp_info.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bmp_info.bmiHeader.biSizeImage   = p_sys->i_atmo_height * p_sys->i_atmo_width * 4;
    bmp_info.bmiHeader.biCompression = BI_RGB;
    bmp_info.bmiHeader.biWidth        = p_sys->i_atmo_width;
    bmp_info.bmiHeader.biHeight       = -p_sys->i_atmo_height;
    bmp_info.bmiHeader.biBitCount     = 32;
    bmp_info.bmiHeader.biPlanes       = 1;

    bmp_fileheader.bfReserved1 = 0;
    bmp_fileheader.bfReserved2 = 0;
    bmp_fileheader.bfSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + bmp_info.bmiHeader.biSizeImage;
    bmp_fileheader.bfType = 'MB';
    bmp_fileheader.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER); 

    fp_bitmap = fopen(psz_filename,"wb");
    if( fp_bitmap != NULL)
    {
        fwrite(&bmp_fileheader, sizeof(BITMAPFILEHEADER), 1, fp_bitmap);
        fwrite(&bmp_info.bmiHeader, sizeof(BITMAPINFOHEADER), 1, fp_bitmap);
        fwrite(p_pixels, bmp_info.bmiHeader.biSizeImage, 1, fp_bitmap); 
        fclose(fp_bitmap);
    }

#endif
}


/*******************************************************************************************
* CreateMiniImage: extracts a 64x48 pixel image from the frame (there is a small border 
* arround thats why the loops starts with one instead zero) without any interpolation 
********************************************************************************************/
static void CreateMiniImage( vout_thread_t *p_vout, picture_t *p_inpic)
{
    vout_sys_t *p_sys = p_vout->p_sys;
    /* pointer to RGB Buffer created in external libary as safe array which is locked inside AtmoLockTransferBuffer */
    uint8_t *p_transfer = NULL;

#if defined( WIN32 )
    /* for debug out only used*/
    char sz_filename[MAX_PATH];
#endif
    
    /*
      Lock the before created VarArray (AtmoCreateTransferBuffers) inside my Wrapper library
      and give me a pointer to the buffer!
      below linux a global buffer may be used and protected with a mutex?
    */
    p_transfer = (uint8_t *)p_sys->pf_AtmoLockTransferBuffer();
    if(p_transfer == NULL) 
    {
       msg_Err( p_vout, "AtmoLight no transferbuffer available. AtmoLight will be disabled!"); 
       p_sys->i_enabled = 0;
       return;
    }

    /* 
      do the call via pointer to function instead of having a case structure here
    */
    p_sys->pf_ExtractMiniImage(p_sys, p_inpic, p_transfer);


#if defined( WIN32 )
    /*
     if debugging enabled save every 128th image to disk
    */
    if((p_sys->b_saveframes == VLC_TRUE) && (p_sys->sz_framepath[0] != 0 ))
    {

       if((p_sys->i_framecounter & 127) == 0) 
       {
          sprintf(sz_filename,"%satmo_dbg_%06d.bmp",p_sys->sz_framepath, p_sys->i_framecounter);
          msg_Dbg(p_vout, "SaveFrame %s",sz_filename);

          SaveBitmap(p_sys, p_transfer, sz_filename);
       }
       p_sys->i_framecounter++;
    }
#endif

   /* show the colors on the wall */
   p_sys->pf_AtmoSendPixelData();
}




/*****************************************************************************
 * Render: calls the extract method and forwards the incomming picture 1:1 
 *****************************************************************************
 * 
 *****************************************************************************/
static void Render( vout_thread_t *p_vout, picture_t *p_pic )
{
    picture_t *p_outpic;
    vout_sys_t *p_sys = p_vout->p_sys;

    /* This is a new frame. Get a structure from the video_output. */
    while( ( p_outpic = vout_CreatePicture( p_sys->p_vout, 0, 0, 0 ) ) == NULL )
    {
        if( p_vout->b_die || p_vout->b_error )
        {
            return;
        }
        msleep( VOUT_OUTMEM_SLEEP );
    }
    vout_DatePicture( p_sys->p_vout, p_outpic, p_pic->date );
    if((p_sys->i_enabled == 1) && (p_sys->pf_ExtractMiniImage != NULL) && (p_sys->b_pause_live == VLC_FALSE))
    {
       CreateMiniImage(p_vout, p_pic);
    }
    vout_CopyPicture( p_vout, p_outpic, p_pic );

    vout_DisplayPicture( p_sys->p_vout, p_outpic );
}

/*****************************************************************************
 * Control: control facility for the vout (forwards to child vout)
 *****************************************************************************/
static int Control( vout_thread_t *p_vout, int i_query, va_list args )
{
    return vout_vaControl( p_vout->p_sys->p_vout, i_query, args );
}

/*****************************************************************************
 * SendEvents: forward mouse and keyboard events to the parent p_vout
 *****************************************************************************/
static int SendEvents( vlc_object_t *p_this, char const *psz_var,
                       vlc_value_t oldval, vlc_value_t newval, void *p_data )
{
    var_Set( (vlc_object_t *)p_data, psz_var, newval );
    return VLC_SUCCESS;
}

/*****************************************************************************
 * SendEventsToChild: forward events to the child/children vout
 *****************************************************************************/
static int SendEventsToChild( vlc_object_t *p_this, char const *psz_var,
                       vlc_value_t oldval, vlc_value_t newval, void *p_data )
{
    vout_thread_t *p_vout = (vout_thread_t *)p_this;
    var_Set( p_vout->p_sys->p_vout, psz_var, newval );
    return VLC_SUCCESS;
}

/*****************************************************************************
 * FadeToColorThread: Threadmethod which changes slowly the color if output is paused
 *****************************************************************************

*****************************************************************************/
static void FadeToColorThread(fadethread_t *p_fadethread ) 
{ 
  vout_sys_t *p_sys = (vout_sys_t *)p_fadethread->p_sys;
  int i_steps_done = 0;
  int i_index;
  int i_pause_red;
  int i_pause_green;
  int i_pause_blue;

  int i_src_red;
  int i_src_green;
  int i_src_blue;

  vlc_thread_ready( p_fadethread ); 

  uint8_t *p_source = NULL;

  /* initialize AtmoWin COM for this thread! */
  p_sys->pf_AtmoInitialize();

  uint8_t *p_transfer = p_sys->pf_AtmoLockTransferBuffer();  
  if(p_transfer != NULL) {
     /* safe colors as "32bit" Integers to avoid overflows*/  
     i_pause_red   = p_sys->ui_pausecolor_red;
     i_pause_blue  = p_sys->ui_pausecolor_blue;
     i_pause_green = p_sys->ui_pausecolor_green;

     /* allocate a temporary buffer for the last send image size less then 15kb */ 
     int i_size = 4 * p_sys->i_atmo_width * p_sys->i_atmo_height;
     p_source = (uint8_t *)malloc( i_size );
     if(p_source != NULL)
     {
        /* get a copy of the last transfered image as orign for the fading steps...*/
        memcpy(p_source, p_transfer, i_size);
     }
     /* send the same pixel data again... to unlock the buffer! */
     p_sys->pf_AtmoSendPixelData(); 

     while( (!p_fadethread->b_die) && (i_steps_done < p_sys->i_fadesteps) && (p_source != NULL)) 
     {
        p_transfer = p_sys->pf_AtmoLockTransferBuffer();  
        if(!p_transfer) break; // should not happen
        i_steps_done++;
        /* 
           move all pixels in the mini image (64x48) one step closer to the desired color 
           these loop takes the most time of this thread improvements wellcome!
        */
        for(i_index = 0; (i_index < i_size) && (!p_fadethread->b_die);  i_index+=4) 
        {
            i_src_blue  = p_source[i_index+0];
            i_src_green = p_source[i_index+1];
            i_src_red   = p_source[i_index+2];
            p_transfer[i_index+0] = (uint8_t) ((((i_pause_blue  - i_src_blue) *i_steps_done)/p_sys->i_fadesteps) + i_src_blue); 
            p_transfer[i_index+1] = (uint8_t) ((((i_pause_green - i_src_green)*i_steps_done)/p_sys->i_fadesteps) + i_src_green); 
            p_transfer[i_index+2] = (uint8_t) ((((i_pause_red   - i_src_red)  *i_steps_done)/p_sys->i_fadesteps) + i_src_red); 
        }

        //sprintf(sz_filename,"%satmo_fade_%06d.bmp",p_sys->sz_framepath, i_steps_done);
        //SaveBitmap(p_sys, p_transfer, sz_filename);

        /* send image to lightcontroller */
        p_sys->pf_AtmoSendPixelData();  
        /* is there something like and interruptable sleep inside the VLC libaries?
           inside native win32 I would use an Event (CreateEvent) and here an WaitForSingleObject?
        */ 
        if(p_fadethread->b_die) break;
        msleep(10000);
        if(p_fadethread->b_die) break;
        msleep(10000);
        if(p_fadethread->b_die) break;
        msleep(10000);
        if(p_fadethread->b_die) break;
        msleep(10000);
     }
     if(p_source != NULL)
        free(p_source);  
  }
  /* call indirect to OleUnitialize() for this thread */
  p_sys->pf_AtmoFinalize(0);

  msg_Dbg(p_sys->p_vout, "Pause color fading thread done");

/*
  vlc_mutex_lock( &p_sys->filter_lock );
  if(p_fadethread->b_die == VLC_FALSE) {
     p_sys->p_fadethread = NULL;
#ifdef WIN32
     CloseHandle( p_fadethread->thread_id );
#endif
     vlc_object_destroy( p_fadethread );
  }
  vlc_mutex_unlock( &p_sys->filter_lock );
  */
}

/******************************************************************************
* CheckAndStopFadeThread: if there is a fadethread structure left, or running...
*******************************************************************************
* this function will stop the thread ... and waits for its termination
* before removeing the objects from vout_sys_t ...
*******************************************************************************/
static void CheckAndStopFadeThread(vout_thread_t *p_vout) 
{
   vout_sys_t *p_sys = (vout_sys_t *)p_vout->p_sys;
   if(p_sys->p_fadethread != NULL)
   {
       vlc_mutex_lock( &p_sys->filter_lock );
       if(p_sys->p_fadethread != NULL) 
       {
           msg_Dbg(p_vout, "kill still running fadeing thread...");

           p_sys->p_fadethread->b_die = VLC_TRUE;

           vlc_thread_join(p_sys->p_fadethread);

#ifdef WIN32
           CloseHandle( p_sys->p_fadethread->thread_id );
#endif
           vlc_object_destroy(p_sys->p_fadethread);
           p_sys->p_fadethread = NULL;
       }
       vlc_mutex_unlock( &p_sys->filter_lock );
   }
}

/*****************************************************************************
 * StateCallback: Callback for the inputs variable "State" to get notified
 * about Pause and Continue Playback events.  
******************************************************************************/
static int StateCallback( vlc_object_t *p_this, char const *psz_cmd,
                          vlc_value_t oldval, vlc_value_t newval,
                          void *p_data )
{
    vout_thread_t *p_vout = (vout_thread_t *)p_data; 
    vout_sys_t *p_sys = (vout_sys_t *)p_vout->p_sys;


    if((p_sys->b_usepausecolor == VLC_TRUE) && (p_sys->i_enabled == 1))
    {
      msg_Dbg(p_vout, "state change from: %d to %d", oldval.i_int, newval.i_int);

      if((newval.i_int == PAUSE_S) && (oldval.i_int == PLAYING_S)) 
      {
        /* tell the other thread to stop sending images to light controller */
        p_sys->b_pause_live = VLC_TRUE;

        // ggf. alten Thread abrumen should not happen....
        CheckAndStopFadeThread(p_vout);

        // perpare spawn fadeing thread
        vlc_mutex_lock( &p_sys->filter_lock );
        /* launch only a new thread if there is none active! or waiting for cleanup */
        if(p_sys->p_fadethread == NULL) 
        { 
            p_sys->p_fadethread = (fadethread_t *)vlc_object_create( p_vout, sizeof(fadethread_t) );
            p_sys->p_fadethread->p_sys = p_sys; 

            if( vlc_thread_create( p_sys->p_fadethread, "AtmoLight fadeing", FadeToColorThread, VLC_THREAD_PRIORITY_LOW, VLC_FALSE ) )
            {
              msg_Err( p_vout, "cannot create FadeToColorThread" );
              vlc_object_destroy( p_sys->p_fadethread );
              p_sys->p_fadethread = NULL;
            }
        }
        vlc_mutex_unlock( &p_sys->filter_lock );
      }

      if((newval.i_int == PLAYING_S) && (oldval.i_int == PAUSE_S)) 
      {
        /* playback continues check thread state */   
        CheckAndStopFadeThread(p_vout);
        /* reactivate the Render function... to do its normal work */
        p_sys->b_pause_live = VLC_FALSE;
      }
    }

    return VLC_SUCCESS;
}
